#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <linux/parport.h>
#include <linux/ppdev.h>

#define NUM_CHANNEL 16
#define BUFFERSIZE 0x408

unsigned char read_buffer[BUFFERSIZE], write_buffer[NUM_CHANNEL][BUFFERSIZE];
unsigned pos[NUM_CHANNEL], read_pos;
unsigned len[NUM_CHANNEL], read_len;
int sock[NUM_CHANNEL];
struct pollfd pollfd[NUM_CHANNEL];
int ppdev = -1;
int echo_pending;
int ret_fork = 0;

static void error(int, const char *);
static void warn(int, const char *);
static void init_sockets(void);
static void cleanup_sockets(void);
static void read_from_socket(int);
static void init_ppdev(void);
static void cleanup_ppdev(void);
static void read_from_ppdev(void);
static void write_to_ppdev(char *buffer, int len);

unsigned char id[] =
"\0\205"
"MANUFACTURER:Hewlett-Packard;"
"COMMAND SET:HP GDI;"
"MODEL:HP LaserJet 3100;"
"CLASS:PRINTER;"
"DESCRIPTION:Hewlett-Packard LaserJet 3100 MFP;";

unsigned char init[] = {0x5a, 0x00, 0x01, 0x00,
			0x00, 0x00, 0x00, 0x00};
unsigned char echo[] = {0x5a, 0x00, 0x00, 0x00,
			0x04, 0x00, 0x00, 0x00,
			'E', 'c', 'h', 'o'};

int main(int argc, char **argv)
{
	int i, ret, arg;

	init_sockets();
	init_ppdev();

	ret_fork = fork();
	error(ret_fork, "fork");
	if (ret_fork)
		exit(0);

	for(;;) {
		read_from_ppdev();

		ret = poll(pollfd, NUM_CHANNEL, 100);
		error(ret, "poll");

		if (ret == 0)
			continue;

		for (i = 0; i < NUM_CHANNEL; i++) {

			if (!(pollfd[i].revents & POLLIN))
				continue;

			read_from_socket(i);
		}
	}

	return 0;
}

void warn(int ret, const char *msg) {
	char msgbuf[256];

	if (ret >= 0 || errno == EPIPE)
		return;

	sprintf(msgbuf, "lj3100_mux: %s", msg);
	perror(msgbuf);
}

void error(int ret, const char *msg) {
	char msgbuf[256];

	if (ret >= 0 || errno == EPIPE)
		return;

	sprintf(msgbuf, "lj3100_mux: %s", msg);
	perror(msgbuf);
	exit(1);
}

static int read_packet(int fd, unsigned char *buffer, unsigned *pos, unsigned *len)
{
	int ret;

	if (*len == *pos) {
		*len = 8;
		*pos = 0;
	}

	ret = read(fd, buffer + *pos, *len - *pos);
	if (ret <= 0)
		return ret;

	*pos += ret;

	if (*pos == 8) {
		*len = 0x100 * buffer[5] + buffer[4] + 8;

		if (*len > BUFFERSIZE) {
			fprintf(stderr, "Buffer overflow %2.2x%2.2x%2.2x%2.2x %2.2x%2.2x%2.2x%2.2x\n",
				buffer[0], buffer[1],
				buffer[2], buffer[3], 
				buffer[4], buffer[5], 
				buffer[6], buffer[7]);
			return -2;
		}
	}

	return ret;
}

static void init_sockets(void)
{
	int i, ret, arg;

	atexit(cleanup_sockets);

	signal(SIGPIPE, SIG_IGN);

	for (i = 0; i < NUM_CHANNEL; i++) {
		struct sockaddr_un sa;

		sa.sun_family = AF_UNIX;
		sprintf(sa.sun_path, "/tmp/lj3100_%d", i);

		remove(sa.sun_path);

		sock[i] = socket(PF_UNIX, SOCK_STREAM, 0);
		error(ret, "socket");

		ret = bind(sock[i], &sa, sizeof(sa));
		error(ret, "bind");

		ret = chmod(sa.sun_path, 0777);
		error(ret, "chmod");

		ret = listen(sock[i], 0);
		error(ret, "listen");

		pollfd[i].fd = sock[i];
		pollfd[i].events = POLLIN;
	}
}

static void reset_socket(int i)
{
	close(pollfd[i].fd);
	pollfd[i].fd = sock[i];
}

static void cleanup_sockets()
{
	int i, ret, arg;

	if (ret_fork != 0)
		return;

	for (i = 0; i < NUM_CHANNEL; i++) {
		char path[50];

		sprintf(path, "/tmp/lj3100_%d", i);

		ret = remove(path);
		error(ret, "remove");
	}
}

static void read_from_socket(int i)
{
	int ret;
	unsigned char status;

	if (pollfd[i].fd == sock[i]) {
		pollfd[i].fd = accept(sock[i], NULL, 0);
		if (pollfd[i].fd < 0) {
			error(ret, "accept");
			pollfd[i].fd = sock[i];
		}
		pos[i] = 0;
		len[i] = 8;
		return;
	}

	ret = read_packet(pollfd[i].fd, write_buffer[i], &pos[i], &len[i]);
	warn(ret, "read_packet");
	if (ret <= 0) {
		reset_socket(i);
		return;
	}

	if (pos[i] < len[i])
		return;

	write_to_ppdev(write_buffer[i], len[i]);
}

static void write_to_socket(void)
{
	int i, ret;

	i = read_buffer[1] >> 4;

	if (pollfd[i].fd == sock[i]) {
		fprintf(stderr, "Dropping packet %2.2x%2.2x%2.2x%2.2x %2.2x%2.2x%2.2x%2.2x %2.2x%2.2x%2.2x%2.2x %2.2x%2.2x%2.2x%2.2x\n",
			read_buffer[0], read_buffer[1],
			read_buffer[2], read_buffer[3],
			read_buffer[4], read_buffer[5], 
			read_buffer[6], read_buffer[7],
			read_buffer[8], read_buffer[9],
			read_buffer[10], read_buffer[11], 
			read_buffer[12], read_buffer[13], 
			read_buffer[14], read_buffer[15]);
		return;
	}

	ret = write(pollfd[i].fd, read_buffer, read_len);
	warn(ret, "write");
	if (ret <= 0)
		reset_socket(i);
}

static void init_ppdev(void)
{
	int ret, arg;
	unsigned char status;

	ppdev = open("/dev/parports/0", O_RDWR);
	error(ppdev, "open /dev/parports/0");

	ret = ioctl(ppdev, PPEXCL);
	error(ret, "ioctl PPEXCL");

	ret = ioctl(ppdev, PPCLAIM);
	error(ret, "ioctl PPCLAIM");

	atexit(cleanup_ppdev);

	arg = IEEE1284_MODE_NIBBLE | IEEE1284_DEVICEID;
	ret = ioctl(ppdev, PPNEGOT, &arg);
	error(ret, "ioctl PPNEGOT IEEE1284_MODE_NIBBLE | IEEE1284_DEVICEID");

	ret = read(ppdev, read_buffer, sizeof(id));
	error(ret, "read id");

	if (memcmp(read_buffer, id, sizeof(id))) {
		fprintf(stderr, "Wrong id\n");
		exit(1);
	}

	arg = IEEE1284_MODE_ECP;
	ret = ioctl(ppdev, PPNEGOT, &arg);
	error(ret, "ioctl PPNEGOT IEEE1284_MODE_ECP");
}

static void cleanup_ppdev(void)
{
	int ret, arg;

	if (ret_fork != 0)
		return;

	if (ppdev < 0)
		return;

	arg = IEEE1284_MODE_COMPAT;
	ret = ioctl(ppdev, PPNEGOT, &arg);
	error(ret, "ioctl PPNEGOT IEEE1284_MODE_COMPAT");

	ret = ioctl(ppdev, PPRELEASE);
	error(ret, "ioctl PPRELEASE");
}

static void read_from_ppdev(void)
{
	int ret, arg;
	unsigned char status;

	for (;;) {
		ret = ioctl(ppdev, PPRSTATUS, &status);
		error(ret, "ioctl PPRSTATUS");

		if (status & PARPORT_STATUS_ERROR)
			return;

		read_pos = 0;
		read_len = 8;

		do {
			ret = read_packet(ppdev, read_buffer, &read_pos, &read_len);
		} while(ret > 0 && read_pos < read_len);

		arg = IEEE1284_MODE_COMPAT;
		ret = ioctl(ppdev, PPNEGOT, &arg);
		error(ret, "ioctl PPNEGOT IEEE1284_MODE_COMPAT");

		arg = IEEE1284_MODE_ECP;
		ret = ioctl(ppdev, PPNEGOT, &arg);
		error(ret, "ioctl PPNEGOT IEEE1284_MODE_COMPAT");

		if (echo_pending &&
		    !memcmp(read_buffer, echo, 6) &&
		    !memcmp(read_buffer + 8, echo + 8, sizeof(echo) - 8)) {
			echo_pending = 0;
			continue;
		}

		if (read_pos == read_len)
			write_to_socket();
	}
}

static void write_to_ppdev(char *buffer, int len)
{
	int ret, arg;
	unsigned char status;

	read_from_ppdev();

	while (len > 0) {
		ret = write(ppdev, buffer, 1);

		if (ret < 0) {
			perror("write to ppdev");
			break;
		}

		buffer += ret;
		len -= ret;
	}

	read_from_ppdev();
}
